import { declare } from '@babel/helper-plugin-utils';
import syntaxExplicitResourceManagement from '@babel/plugin-syntax-explicit-resource-management';
import { traverse, types, template } from '@babel/core';

var index = declare(api => {
  api.assertVersion("8.0.0-alpha.5");
  const TOP_LEVEL_USING = new Map();
  function isUsingDeclaration(node) {
    if (!types.isVariableDeclaration(node)) return false;
    return node.kind === "using" || node.kind === "await using" || TOP_LEVEL_USING.has(node);
  }
  const transformUsingDeclarationsVisitor = {
    ForOfStatement(path) {
      const {
        left
      } = path.node;
      if (!isUsingDeclaration(left)) return;
      const {
        id
      } = left.declarations[0];
      const tmpId = path.scope.generateUidIdentifierBasedOnNode(id);
      left.declarations[0].id = tmpId;
      left.kind = "const";
      path.ensureBlock();
      path.node.body.body.unshift(types.variableDeclaration("using", [types.variableDeclarator(id, types.cloneNode(tmpId))]));
    },
    "BlockStatement|StaticBlock"(path, state) {
      let stackId = null;
      let needsAwait = false;
      for (const node of path.node.body) {
        if (!isUsingDeclaration(node)) continue;
        stackId ??= path.scope.generateUidIdentifier("stack");
        const isAwaitUsing = node.kind === "await using" || TOP_LEVEL_USING.get(node) === 1;
        needsAwait ||= isAwaitUsing;
        if (!TOP_LEVEL_USING.delete(node)) {
          node.kind = "const";
        }
        node.declarations.forEach(decl => {
          const args = [types.cloneNode(stackId), decl.init];
          if (isAwaitUsing) args.push(types.booleanLiteral(true));
          decl.init = types.callExpression(state.addHelper("using"), args);
        });
      }
      if (!stackId) return;
      const errorId = path.scope.generateUidIdentifier("error");
      const hasErrorId = path.scope.generateUidIdentifier("hasError");
      let disposeCall = types.callExpression(state.addHelper("dispose"), [types.cloneNode(stackId), types.cloneNode(errorId), types.cloneNode(hasErrorId)]);
      if (needsAwait) disposeCall = types.awaitExpression(disposeCall);
      const replacement = template.statement.ast`
        try {
          var ${stackId} = [];
          ${path.node.body}
        } catch (_) {
          var ${errorId} = _;
          var ${hasErrorId} = true;
        } finally {
          ${disposeCall}
        }
      `;
      types.inherits(replacement.block, path.node);
      const {
        parentPath
      } = path;
      if (parentPath.isFunction() || parentPath.isTryStatement() || parentPath.isCatchClause()) {
        path.replaceWith(types.blockStatement([replacement]));
      } else if (path.isStaticBlock()) {
        path.node.body = [replacement];
      } else {
        path.replaceWith(replacement);
      }
    }
  };
  const transformUsingDeclarationsVisitorSkipFn = traverse.visitors.merge([transformUsingDeclarationsVisitor, {
    Function(path) {
      path.skip();
    }
  }]);
  return {
    name: "proposal-explicit-resource-management",
    inherits: syntaxExplicitResourceManagement,
    visitor: traverse.visitors.merge([transformUsingDeclarationsVisitor, {
      Program(path) {
        TOP_LEVEL_USING.clear();
        if (path.node.sourceType !== "module") return;
        if (!path.node.body.some(isUsingDeclaration)) return;
        const innerBlockBody = [];
        for (const stmt of path.get("body")) {
          if (stmt.isFunctionDeclaration() || stmt.isImportDeclaration()) {
            continue;
          }
          let {
            node
          } = stmt;
          let shouldRemove = true;
          if (stmt.isExportDefaultDeclaration()) {
            let {
              declaration
            } = stmt.node;
            let varId;
            if (types.isClassDeclaration(declaration)) {
              varId = declaration.id;
              declaration.id = null;
              declaration = types.toExpression(declaration);
            } else if (!types.isExpression(declaration)) {
              continue;
            }
            varId ??= path.scope.generateUidIdentifier("_default");
            innerBlockBody.push(types.variableDeclaration("var", [types.variableDeclarator(varId, declaration)]));
            stmt.replaceWith(types.exportNamedDeclaration(null, [types.exportSpecifier(types.cloneNode(varId), types.identifier("default"))]));
            continue;
          }
          if (stmt.isExportNamedDeclaration()) {
            node = stmt.node.declaration;
            if (!node || types.isFunction(node)) continue;
            stmt.replaceWith(types.exportNamedDeclaration(null, Object.keys(types.getOuterBindingIdentifiers(node, false)).map(id => types.exportSpecifier(types.identifier(id), types.identifier(id)))));
            shouldRemove = false;
          } else if (stmt.isExportDeclaration()) {
            continue;
          }
          if (types.isClassDeclaration(node)) {
            const {
              id
            } = node;
            node.id = null;
            innerBlockBody.push(types.variableDeclaration("var", [types.variableDeclarator(id, types.toExpression(node))]));
          } else if (types.isVariableDeclaration(node)) {
            if (node.kind === "using") {
              TOP_LEVEL_USING.set(stmt.node, 0);
            } else if (node.kind === "await using") {
              TOP_LEVEL_USING.set(stmt.node, 1);
            }
            node.kind = "var";
            innerBlockBody.push(node);
          } else {
            innerBlockBody.push(stmt.node);
          }
          if (shouldRemove) stmt.remove();
        }
        path.pushContainer("body", types.blockStatement(innerBlockBody));
      },
      Function(path, state) {
        if (path.node.async) {
          path.traverse(transformUsingDeclarationsVisitorSkipFn, state);
        }
      }
    }])
  };
});

export { index as default };
//# sourceMappingURL=index.js.map
